﻿// ---------------------------------------------------------------------------
// 拡張子「.tpl」の拡張スクリプト
// 
// 概要
//   タグ <% statement %> の statement を、文として実行します。
//   タグ <%$ expression %> の expression を、式として出力します。（組み込み関数 print( expression ) に置換される）
//   ※対応言語は C# のみです。
// 
// 組み込み変数
//   string    __MASHIE__   MashIEのファイル名
//   string    __PLUGIN__   このファイル名
//   string    __FILE__     tplのファイル名
//   Hashtable _IN          tplに渡されたパラメータ（HTMLのアンカーやフォーム、組み込み関数 include から渡される）
// 
// 組み込み関数
//   void print ( object mesasage )							message を文字列で出力します。
//   void print_r ( object mesasage )						message を文字列で出力します。（Json形式に整形される）
//   string include ( string fileName )						fileName を tpl として実行し、出力を文字列で取得します。
//   string include ( string fileName , Hashtable args )	fileName を tpl として実行し、出力を文字列で取得します。（args がパラメータとして渡される）
// ---------------------------------------------------------------------------

#pragma link "mash.exe"
using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Text;

// tplクラス
public class tpl
{
	// キャッシュクラス
	public class Cache
	{
		// フィールド
		public string   fileName      = "";				// ファイル名
		public DateTime dateTime      = new DateTime();	// 日時
		public bool     coded         = false;			// コードフラグ
		public string   source        = "";				// ソース
		public int[]    sourceLineNos = null;			// ソース行番号リスト
		public string   className     = "";				// クラス名
		public string   methodName    = "";				// メソッド名
		public string[] literals      = null;			// リテラルリスト
	}
	
	// スタティックフィールド
	public static bool debug    = false;		// デバッグフラグ(trueにすると解析結果のソースが出力される)
	public static tpl  instance = new tpl();	// インスタンス
	
	// スタティックメソッド
	public static void Main(string[] args)
	{
		// 引数を取得
		string    mashieFileName = args[0];
		string    pluginFileName = args[1];
		string    targetFileName = args[2];
		Hashtable targetFileArgs = new Hashtable();
		for (int i = 3; i < args.Length; i ++)
		{
			string arg = args[i];
			if (arg == "")
			{
				continue;
			}
			string key = "";
			string val = "";
			int pos = arg.IndexOf("=");
			if (pos >= 0)
			{
				key = arg.Substring(0, pos);
				val = arg.Substring(pos + 1);
			}
			else
			{
				key = arg;
			}
			if (key == "")
			{
				continue;
			}
			
			object targetFileArg = targetFileArgs[key];
			if (targetFileArg is ArrayList)
			{
				((ArrayList)targetFileArg).Add(val);
			}
			else if (targetFileArg is string)
			{
				targetFileArg = new ArrayList(){targetFileArg, val};
			}
			else
			{
				targetFileArg = val;
			}
			targetFileArgs[key] = targetFileArg;
		}
		
		// 解析結果を出力
		Console.Write(instance.include(mashieFileName, pluginFileName, targetFileName, targetFileArgs));
	}
	
	// フィールド
	private string                    _prefixSource    = "";								// 前半ソース
	private int                       _prefixLineCount = 0;									// 前半行数
	private List<int>                 _prefixLineNos   = null;								// 前半行番号リスト
	private string                    _surfixSource    = "";								// 後半ソース
	private int                       _surfixLineCount = 0;									// 後半行数
	private Dictionary<string, Cache> _caches          = new Dictionary<string, Cache>();	// キャッシュ
	
	// コンストラクタ
	public tpl()
	{
		// 前半ソースを作成
		StringBuilder prefixSource = new StringBuilder();
		prefixSource.Append("#pragma import \"$PLUGIN_FILE\"").Append("\n");
		prefixSource.Append("public class $CLASS_NAME").Append("\n");
		prefixSource.Append("{").Append("\n");
		prefixSource.Append("	public static string $METHOD_NAME(string mashieFileName, string pluginFileName, string targetFileName, System.Collections.Hashtable targetFileArgs, string[] literals)").Append("\n");
		prefixSource.Append("	{").Append("\n");
		prefixSource.Append("		$CLASS_NAME instance = new $CLASS_NAME();").Append("\n");
		prefixSource.Append("		instance.__MASHIE__  = mashieFileName;").Append("\n");
		prefixSource.Append("		instance.__PLUGIN__  = pluginFileName;").Append("\n");
		prefixSource.Append("		instance.__FILE__    = targetFileName;").Append("\n");
		prefixSource.Append("		instance.__LITERAL__ = literals;").Append("\n");
		prefixSource.Append("		instance._OUT        = new System.Text.StringBuilder();").Append("\n");
		prefixSource.Append("		instance._IN         = targetFileArgs;").Append("\n");
		prefixSource.Append("		instance.inner$METHOD_NAME();").Append("\n");
		prefixSource.Append("		return instance._OUT.ToString();").Append("\n");
		prefixSource.Append("	}").Append("\n");
		prefixSource.Append("	public string                       __MASHIE__  = \"\";").Append("\n");
		prefixSource.Append("	public string                       __PLUGIN__  = \"\";").Append("\n");
		prefixSource.Append("	public string                       __FILE__    = \"\";").Append("\n");
		prefixSource.Append("	public string[]                     __LITERAL__ = null;").Append("\n");
		prefixSource.Append("	public System.Text.StringBuilder    _OUT        = null;").Append("\n");
		prefixSource.Append("	public System.Collections.Hashtable _IN         = null;").Append("\n");
		prefixSource.Append("	private void print(object message)").Append("\n");
		prefixSource.Append("	{").Append("\n");
		prefixSource.Append("		$PLUGIN_NAME.instance.print(_OUT, message);").Append("\n");
		prefixSource.Append("	}").Append("\n");
		prefixSource.Append("	private void print_r(object message)").Append("\n");
		prefixSource.Append("	{").Append("\n");
		prefixSource.Append("		$PLUGIN_NAME.instance.print_r(_OUT, message);").Append("\n");
		prefixSource.Append("	}").Append("\n");
		prefixSource.Append("	private string include(string fileName)").Append("\n");
		prefixSource.Append("	{").Append("\n");
		prefixSource.Append("		return $PLUGIN_NAME.instance.include(__MASHIE__, __PLUGIN__, fileName, null);").Append("\n");
		prefixSource.Append("	}").Append("\n");
		prefixSource.Append("	private string include(string fileName, System.Collections.Hashtable args)").Append("\n");
		prefixSource.Append("	{").Append("\n");
		prefixSource.Append("		return $PLUGIN_NAME.instance.include(__MASHIE__, __PLUGIN__, fileName, args);").Append("\n");
		prefixSource.Append("	}").Append("\n");
		prefixSource.Append("	public void inner$METHOD_NAME()").Append("\n");
		prefixSource.Append("	{").Append("\n");
		
		_prefixSource    = prefixSource.ToString();
		_prefixLineCount = _prefixSource.Length - _prefixSource.Replace("\n", "").Length;
	 	_prefixLineNos   = new List<int>();
		for (int i = 0; i < _prefixLineCount; i ++)
		{
			_prefixLineNos.Add(-_prefixLineCount + i);
		}
		
		// 後半ソースを作成
		StringBuilder surfixSource = new StringBuilder();
		surfixSource.Append("	}").Append("\n");
		surfixSource.Append("}").Append("\n");
		 
		_surfixSource    = surfixSource.ToString();
		_surfixLineCount = _surfixSource.Length - _surfixSource.Replace("\n", "").Length;
	}
	
	// メソッド
	public string include(string mashieFileName, string pluginFileName, string targetFileName, Hashtable targetFileArgs)
	{
		// ファイル名を取得
		try
		{
			targetFileName = Path.GetFullPath(targetFileName);
		}
		catch
		{
		}
		if (!File.Exists(targetFileName))
		{
			return "";
		}
		
		// ファイル日時を取得
		DateTime targetFileDate = File.GetLastWriteTime(targetFileName);
		
		// カレントディレクトリを移動
		string filePath = Environment.CurrentDirectory;
		Environment.CurrentDirectory = Path.GetDirectoryName(targetFileName);
		
		// キャッシュを取得
		Cache cache   = null;
		bool  changed = true;
		if (_caches.ContainsKey(targetFileName))
		{
			cache   = _caches[targetFileName];
			changed = (cache.dateTime != targetFileDate);
		}
		
		// 変更がある場合
		if (changed)
		{
			try
			{
				// 既存のキャッシュを削除
				if (cache != null)
				{
					_caches.Remove(targetFileName);
					cache = null;
				}
				
				// ファイルを読込
				string targetFileText = File.ReadAllText(targetFileName).Replace("\r", "");
				if (targetFileText.Length <= 0)
				{
					return "";
				}
				
				// ソースを構築
				string       pluginName    = Path.GetFileNameWithoutExtension(pluginFileName);
				string       className     = Path.GetFileNameWithoutExtension(targetFileName);
				string       methodName    = "Execute";
				List<int>    sourceLineNos = new List<int>();
				List<string> literals      = new List<string>();
				
				// 前半ソースを作成
				StringBuilder prefixSource = new StringBuilder(_prefixSource);
				prefixSource.Replace("$MASHIE_FILE", mashieFileName.Replace("\\", "\\\\"));
				prefixSource.Replace("$PLUGIN_FILE", pluginFileName.Replace("\\", "\\\\"));
				prefixSource.Replace("$PLUGIN_NAME", pluginName);
				prefixSource.Replace("$CLASS_NAME", className);
				prefixSource.Replace("$METHOD_NAME", methodName);
				foreach (int prefixLineNo in _prefixLineNos)
				{
					sourceLineNos.Add(prefixLineNo);
				}
				
				// 内部ソースを作成
			 	StringBuilder innerSource = new StringBuilder();
				bool          innerCoded  = false;
				
				{
					StringBuilder token         = new StringBuilder();
					int           lineNo        = 1;
					int           literalLineNo = 1;
					bool          coding        = false;
					bool          codePrinting  = false;
					
					char[] cs = targetFileText.ToCharArray();
					int    cn = cs.Length - 1;
					for (int ci = 0; ci <= cn; ci ++)
					{
						bool ends   = (ci == cn);
						char c      = cs[ci];
						char cNext  = ' ';
						char cNext2 = ' ';
						if (ci + 1 <= cn) cNext  = cs[ci + 1];
						if (ci + 2 <= cn) cNext2 = cs[ci + 2];
						
						// コード中の場合
						if (coding)
						{
							// 終了タグの場合
							if (ends || (c == '%' && cNext == '>'))
							{
								// トークンが残っている場合
								if (token.Length > 0)
								{
									// コードありとする
									innerCoded = true;
									
									// コードプリント中の場合
									if (codePrinting)
									{
										// コードソースを追加(トークンをプリント)
										sourceLineNos.Add(lineNo);
										innerSource.Append("print(").Append(token.ToString()).Append(");").Append('\n');
									}
									// コードプリント中でない場合
									else
									{
										// 定義文の場合
										string tokenText = token.ToString();
										if (tokenText.Replace("\t", " ").Trim().StartsWith("using "))
										{
											// 前半ソースの先頭に追加(トークン)
											sourceLineNos.Insert(0, lineNo);
											prefixSource.Insert(0, '\n').Insert(0, tokenText);
										}
										// 定義文でない場合
										else
										{
											// コードソースを追加(トークン)
											sourceLineNos.Add(lineNo);
											innerSource.Append(token.ToString()).Append('\n');
										}
									}
									
									// トークンをクリア
									token = new StringBuilder();
								}
								
								// コード中をOFF
								coding = false;
								
								// コードプリント中をOFF
								codePrinting = false;
								
								// 2文字 %> 処理したので追加で1文字分進めておく
								ci ++;
								
								// リテラルの開始行番号を設定
								literalLineNo = lineNo;
							}
							// 改行の場合
							else if (c == '\n')
							{
								// トークンが残っている場合
								if (token.Length > 0)
								{
									// コードありとする
									innerCoded = true;
									
									// 定義文の場合
									string tokenText = token.ToString().Trim();
									if (tokenText.Replace("\t", " ").Trim().StartsWith("using "))
									{
										// 前半ソースの先頭に追加(トークン)
										sourceLineNos.Insert(0, lineNo);
										prefixSource.Insert(0, '\n').Insert(0, tokenText);
									}
									// 定義文でない場合
									else
									{
										// コードソースを追加(トークン)
										sourceLineNos.Add(lineNo);
										innerSource.Append(token.ToString()).Append('\n');
									}
									
									// トークンをクリア
									token = new StringBuilder();
								}
								
								// 行番号を加算
								lineNo ++;
							}
							// その他の場合
							else
							{
								// トークンを追加
								token.Append(c);
							}
						}
						// コード中でない場合
						else
						{
							// 開始タグの場合
							if (ends || (c == '<' && cNext == '%'))
							{
								// トークンが残っている場合
								if (token.Length > 0)
								{
									// リテラルを追加
									literals.Add(token.ToString());
									
									// コードソースを追加(リテラル)
									sourceLineNos.Add(literalLineNo);
									innerSource.Append("_OUT.Append(__LITERAL__[").Append(literals.Count - 1).Append("]);").Append('\n');
									
									// トークンをクリア
									token = new StringBuilder();
								}
								
								// コード中をOFF
								coding = true;
								
								// プリントタグの場合
								if (cNext2 == '$')
								{
									// コードプリント中をON
									codePrinting = true;
									
									// 3文字 <%$ 処理したので追加で2文字分進めておく
									ci += 2;
								}
								// プリントタグでない場合
								else
								{
									// コードプリント中をOFF
									codePrinting = false;
									
									// 2文字 <% 処理したので追加で1文字分進めておく
									ci ++;
								}
							}
							// 改行の場合
							else if (c == '\n')
							{
								// トークンを追加(文字)
								token.Append(c);
								
								// 行番号を加算
								lineNo ++;
							}
							// その他の場合
							else
							{
								// トークンを追加(文字)
								token.Append(c);
							}
						}
					}
				}
				
				// 後半ソースを作成
				StringBuilder surfixSource = new StringBuilder(_surfixSource);
				int surfixLineOffset = sourceLineNos[sourceLineNos.Count - 1] + 1;
				for (int i = 0; i <= _surfixLineCount; i ++)
				{
					sourceLineNos.Add(surfixLineOffset + i);
				}
				
				// キャッシュを追加
				cache = new Cache();
				cache.fileName = targetFileName;
				cache.dateTime = targetFileDate;
				if (innerCoded)
				{
					cache.coded         = true;
					cache.source        = prefixSource.ToString() + innerSource.ToString() + surfixSource.ToString();
					cache.sourceLineNos = sourceLineNos.ToArray();
					cache.className     = className;
					cache.methodName    = methodName;
					cache.literals      = literals.ToArray();
				}
				else
				{
					cache.coded  = false;
					cache.source = targetFileText;
				}
				_caches[targetFileName] = cache;
				
				// ソースを出力
				if (debug && cache.coded)
				{
					string debugFileName = Path.GetFileName(cache.fileName);
					Console.Write("<br>");
					Console.Write(debugFileName + " ---------------------------------------<br>");
					string[] debugLines       = cache.source.Replace("<", "&lt;").Replace(">", "&gt;").Split('\n');
					int      debugLineCount   = debugLines.Length;
					int[]    debugLineNos     = cache.sourceLineNos;
					int      debugLineNoCount = debugLineNos.Length;
					for (int i = 0; i < debugLineCount; i ++)
					{
						Console.Write((i < debugLineNoCount ? debugLineNos[i].ToString("#,0") : "?") + "：" + debugLines[i] + "<br>");
					}
					Console.Write("--------------------------------------- " + debugFileName + "<br>");
					Console.Write("<br>");
				}
			}
			catch (TargetInvocationException tie)
			{
				throw tie;
			}
			catch (Exception e)
			{
				throw e;
			}
			finally
			{
				// カレントディレクトリを戻す
				Environment.CurrentDirectory = filePath;
			}
		}
		
		// コードがある場合
		if (cache.coded)
		{
			try
			{
				// メソッド引数を作成
				object[] methodArgs = new object[]{mashieFileName, pluginFileName, cache.fileName, (targetFileArgs != null ? targetFileArgs : new Hashtable()), cache.literals};
				
				// ソースを実行して返す
				return mash.CodeUtility.Execute(cache.fileName, cache.source, null, mash.CodeTypes.Auto, cache.className, cache.methodName, methodArgs) + "";
			}
			catch (mash.CodeException ce)
			{
				// 行番号を差し替え
				foreach (mash.CodeError item in ce.errors.items)
				{
					if (item.fileName == cache.fileName)
					{
						int line = item.line;
						if (line > 0 && line <= cache.sourceLineNos.Length)
						{
							line = cache.sourceLineNos[line];
						}
						item.line = line;
					}
				}
				throw ce;
			}
			catch (Exception e)
			{
				throw e;
			}
		}
		// コードがない場合
		else
		{
			// ソースをそのまま返す
			return cache.source;
		}
	}
	
	public void print(StringBuilder sb, object message)
	{
		// メッセージを出力
		if (message != null)
		{
			sb.Append(message);
		}
	}
	
	public void print_r(StringBuilder sb, object message)
	{
		// メッセージを出力
		if (message != null)
		{
			if (message is ArrayList)
			{
				print_r(sb, (ArrayList)message);
			}
			else if (message is Hashtable)
			{
				print_r(sb, (Hashtable)message);
			}
			else
			{
				string text = message.ToString();
				ulong  ulongValue;
				long   longValue;
				double doubleValue;
				bool   boolValue;
				if (ulong.TryParse(text, out ulongValue))
				{
					sb.Append( ulongValue);
				}
				else if (long.TryParse(text, out longValue))
				{
					sb.Append(  longValue);
				}
				else if (double.TryParse(text, out doubleValue))
				{
					sb.Append(doubleValue);
				}
				else if (bool.TryParse(text, out boolValue))
				{
					sb.Append(  boolValue);
				}
				else
				{
					sb.Append("\"");
					sb.Append(message);
					sb.Append("\"");
				}
			}
		}
		else
		{
			sb.Append("null");
		}
	}
	
	public void print_r(StringBuilder sb, ArrayList message)
	{
		// メッセージを出力
		if (message != null)
		{
			sb.Append("[");
			bool isFirst = true;
			for (int i = 0; i < message.Count; i ++)
			{ 
				if (!isFirst) sb.Append(",");
				print_r(sb, message[i]);
				isFirst = false;
			} 
			sb.Append("]");
		}
		else
		{
			sb.Append("null");
		}
	}
	
	public void print_r(StringBuilder sb, Hashtable message)
	{
		// メッセージを出力
		if (message != null)
		{
			sb.Append("{");
			bool isFirst = true;
			foreach (string key in message.Keys)
			{ 
				if (!isFirst) sb.Append(",");
				sb.Append("\"").Append(key).Append("\":");
				print_r(sb, message[key]);
				isFirst = false;
			} 
			sb.Append("}");
		}
		else
		{
			sb.Append("null");
		}
	}
}
